from abc import ABC, abstractmethod


def _iterableify(x):
    if isinstance(x, (str)):
        return [x]
    return x


class LineupConstraints(object):
    def __init__(self,
                 locked: list = [],
                 position_locked: list = [],
                 banned: list = [],
                 position_banned: list = [],
                 groups: list = []):
        self._constraints = []
        self._banned = set()
        self._locked = set()
        self._position_locked = set()
        self._position_banned = set()

        for name in banned:
            self.ban(name)

        for solver_id in position_banned:
            self.position_ban(solver_id)

        for solver_id in position_locked:
            self.position_lock(solver_id)

        for name in locked:
            self.lock(name)

        for players, bounds in groups:
            self.add_group_constraint(players, bounds)

    def exist(self):
        return len(self)

    def __iter__(self):
        return ConstraintIterator(self._constraints)

    def __len__(self):
        return (
            len(self._constraints) +
            len(self._locked) +
            len(self._banned) +
            len(self._position_locked) +
            len(self._position_banned)
        )

    def __repr__(self):
        constraints = ', '.join([repr(c) for c in self._constraints])
        lcs = 'LineupConstraintSet: {}'.format(constraints)
        b1 = '<Banned: {!r}>'.format(self._banned)
        bp1 = '<PositionBanned: {!r}>'.format(self._position_banned)
        lp1 = '<Locked: {!r}>'.format(self._locked)
        l1 = '<PositionLocked: {!r}>'.format(self._position_locked)
        return '<{}, {}, {}, {}, {}>'.format(lcs, b1, bp1, l1, lp1)

    def __str__(self):
        lines = [str(c) for c in self._constraints]
        if len(self._banned):
            lines.append('Banned: ' + ', '.join(p for p in self._banned))
        if len(self._locked):
            lines.append('Locked: ' + ', '.join(p for p in self._locked))
        if len(lines):
            return '\n'.join(lines)
        else:
            return 'None'

    def __eq__(self, constraintset):
        if len(self._constraints) != len(constraintset._constraints):
            return False

        if set(self._constraints) != set(constraintset._constraints):
            return False

        if self._locked != constraintset._locked:
            return False

        if self._banned != constraintset._banned:
            return False

        if self._position_locked != constraintset._position_locked:
            return False

        if self._position_banned != constraintset._position_banned:
            return False

        return True

    def __contains__(self, player):
        if player in self._locked:
            return True

        if player in self._banned:
            return True

        for c in self._constraints:
            if isinstance(c, PlayerGroupConstraint):
                if player in c.players:
                    return True

        return False

    def _check_conflicts(self, constraint):
        if isinstance(constraint, PlayerGroupConstraint):
            for p in constraint.players:
                if p in self._locked or p in self._banned:
                    raise ConstraintConflictException(
                        'Ban/lock constraint for {} already exists'.format(p)
                    )

    def _add(self, constraint):
        self._check_conflicts(constraint)

        if constraint not in self._constraints:
            self._constraints.append(constraint)
        else:
            raise ConstraintConflictException('Duplicate constraint')

    def is_banned(self, player: str) -> bool:
        return player in self._banned

    def is_locked(self, player: str) -> bool:
        return player in self._locked

    def is_position_locked(self, solver_id: str) -> bool:
        return solver_id in self._position_locked

    def is_position_banned(self, solver_id: str) -> bool:
        return solver_id in self._position_banned

    def has_group_constraints(self) -> bool:
        return len(self._constraints) != 0

    def add_group_constraint(self, players, bound):
        self._add(PlayerGroupConstraint(players, bound))

    def ban(self, players):
        _players = _iterableify(players)

        if len(_players) == 0:
            raise ConstraintException('Empty ban group')

        for p in _players:
            if p in self:
                raise ConstraintConflictException(
                    '{} exists in another constraint'.format(p)
                )
        self._banned.update(_players)

    def lock(self, players):
        _players = _iterableify(players)

        if len(_players) == 0:
            raise ConstraintException('Empty lock group')

        for p in _players:
            if p in self:
                raise ConstraintConflictException(
                    '{} exists in another constraint'.format(p)
                )
        self._locked.update(_players)

    def position_lock(self, solver_ids):
        _solver_ids = _iterableify(solver_ids)

        if len(_solver_ids) == 0:
            raise ConstraintException('Empty position lock group')

        for p in _solver_ids:
            if p in self:
                raise ConstraintConflictException(
                    '{} exists in another constraint'.format(p)
                )
        self._position_locked.update(_solver_ids)

    def position_ban(self, solver_ids):
        _solver_ids = _iterableify(solver_ids)

        if len(_solver_ids) == 0:
            raise ConstraintException('Empty position lock group')

        for p in _solver_ids:
            if p in self:
                raise ConstraintConflictException(
                    '{} exists in another constraint'.format(p)
                )
        self._position_banned.update(_solver_ids)


class ConstraintConflictException(Exception):
    pass


class ConstraintIterator(object):
    def __init__(self, constraints):
        self._constraints = constraints
        self.ndx = 0

    def __next__(self):
        if self.ndx >= len(self._constraints):
            raise StopIteration

        r = self._constraints[self.ndx]
        self.ndx += 1
        return r


class AbstractConstraint(ABC):
    @abstractmethod
    def __init__(self):
        super().__init__()

    @abstractmethod
    def __repr__(self):
        pass

    @abstractmethod
    def __str__(self):
        pass

    @abstractmethod
    def __eq__(self, constraint):
        pass

    @abstractmethod
    def __hash__(self):
        pass

    @abstractmethod
    def __contains__(self, player):
        pass


class ConstraintException(Exception):
    pass


class PlayerConstraint(AbstractConstraint):
    def __init__(self, players):
        if not len(players):
            raise ConstraintException('No players in group')

        if len(players) != len(set(players)):
            raise ConstraintException('Duplicate players in group')

        self.players = players

        super().__init__()

    def __eq__(self, rule):
        return set(self.players) == set(rule.players)

    def __hash__(self):
        return hash(''.join(sorted(self.players)))

    def __contains__(self, player):
        return player in self.players


class PlayerGroupConstraint(PlayerConstraint):
    def __init__(self, players, bound):
        super().__init__(players)
        self.exact = None
        self.lb = None
        self.ub = None

        if isinstance(bound, (list, tuple)) and len(bound) == 2:
            self.lb = bound[0]
            self.ub = bound[1]
            self._ub_lb_bounds_sanity_check()
        elif isinstance(bound, int):
            self.exact = bound
            self._exact_bounds_sanity_check()
        else:
            raise ConstraintException('Bound must be length 2 or int')

    def __repr__(self):
        return '<PlayerGroupConstraint: {} of {}>'.format(self._bounds_str,
                                                          self.players)

    def __str__(self):
        return 'Using {} of: {}'.format(
            self._bounds_str,
            ', '.join(p for p in self.players)
        )

    def __eq__(self, constraint):
        return super().__eq__(constraint) and self.exact == constraint.exact \
               and self.lb == constraint.lb and self.ub == constraint.ub

    def __hash__(self):
        return hash((super().__hash__(), self.exact, self.lb, self.ub))

    @property
    def _bounds_str(self):
        if self.exact:
            return '{0.exact}'.format(self)

        return '{0.lb} to {0.ub}'.format(self)

    def _exact_bounds_sanity_check(self):
        if self.exact <= 0:
            raise ConstraintException(
                'Exact bound may not less than or equal to zero'
            )
        if self.exact >= len(self.players):
            raise ConstraintException(
                'Exact bound may not be greater than or equal to number '
                'of players in group'
            )

    def _ub_lb_bounds_sanity_check(self):
        if self.lb < 1:
            raise ConstraintException(
                'Lower bound for {!r} cannot be less than 1'.format(self)
            )
        if self.ub == self.lb:
            raise ConstraintException(
                'Lower bound for {!r} cannot equal upper bound'.format(self)
            )
        if self.ub < self.lb:
            raise ConstraintException(
                'Upper bound for {!r} cannot be less than lower bound.'
                .format(self)
            )
        if self.ub > len(self.players) or self.lb > len(self.players):
            raise ConstraintException(
                'Bound for {!r} cannot be greater than number of players '
                'group'.format(self)
            )



DRAFT_KINGS = 'DRAFT_KINGS'
FAN_DUEL = 'FAN_DUEL'


ROSTER_SIZE = {
    DRAFT_KINGS: {
        'NFL': 9,
        'NFL_SHOWDOWN': 6,
        'MLB_SHOWDOWN': 6,
        'NBA': 8,
        'NBA_SHOWDOWN': 6,
        'WNBA': 6,
        'MLB': 10,
        'SOCCER': 8,
        'EL': 6,
        'NHL': 9,
        'NHL_SHOWDOWN': 6,
        'XFL': 7,
        'TEN': 6,
        'PGA': 6,
        'CSGO_SHOWDOWN': 6,
    },
    FAN_DUEL: {
        'NFL': 9,
        'NFL_MVP': 5,
        'NBA': 9,
        'MLB': 9,
        'WNBA': 7,
        'NASCAR': 5,
        'PGA': 6,
    }
}


SALARY_CAP = {
    DRAFT_KINGS: {
        'NFL': 50_000,
        'NFL_SHOWDOWN': 50_000,
        'NBA': 50_000,
        'NBA_SHOWDOWN': 50_000,
        'WNBA': 50_000,
        'MLB': 50_000,
        'SOCCER': 50_000,
        'EL': 50_000,
        'NHL': 50_000,
        'NHL_SHOWDOWN': 50_000,
        'MLB_SHOWDOWN': 50_000,
        'XFL': 50_000,
        'TEN': 50_000,
        'PGA': 50_000,
        'CSGO_SHOWDOWN': 50_000,
    },
    FAN_DUEL: {
        'NFL': 60_000,
        'NFL_MVP': 60_000,
        'NBA': 60_000,
        'MLB': 35_000,
        'WNBA': 40_000,
        'NASCAR': 50_000,
        'PGA': 60_000,
    }
}


def get_nfl_positions(
    rb_min=2,
    wr_min=3,
    te_min=1,
    te_upper=2,
    d_abbrev='DST',
):
    return [
        ['QB', 1, 1],
        ['RB', rb_min, 3],
        ['WR', wr_min, 4],
        ['TE', te_min, te_upper],
        [d_abbrev, 1, 1]
    ]


def get_nfl_showdown_positions(dk: bool = False, fd: bool = False) -> list:
    if dk:
        ub = 5
    elif fd:
        ub = 4
    else:
        raise NotImplementedError

    return [
        ['CPT', 1, 1],
        ['FLEX', ub, ub]
    ]


POSITIONS = {
    DRAFT_KINGS: {
        'NBA': [
            ['PG', 1, 3],
            ['SG', 1, 3],
            ['SF', 1, 3],
            ['PF', 1, 3],
            ['C', 1, 2]
        ],
        'NBA_SHOWDOWN': [
            ['CPT', 1, 1],
            ['FLEX', 5, 5],
        ],
        'WNBA': [
            ['PG', 1, 3],
            ['SG', 1, 3],
            ['SF', 1, 4],
            ['PF', 1, 4],
        ],
        'NFL': get_nfl_positions(),
        'NFL_SHOWDOWN': get_nfl_showdown_positions(dk=True),
        'MLB': [
            ['SP', 0, 2],
            ['RP', 0, 2],
            ['C', 1, 1],
            ['1B', 1, 1],
            ['2B', 1, 1],
            ['3B', 1, 1],
            ['SS', 1, 1],
            ['OF', 3, 3],
        ],
        'SOCCER': [
            ['F', 2, 3],
            ['M', 2, 3],
            ['D', 2, 3],
            ['GK', 1, 2],
        ],
        'EL': [
            ['G', 2, 3],
            ['F', 3, 4],
        ],
        'NHL': [
            ['C', 2, 3],
            ['W', 3, 4],
            ['D', 2, 3],
            ['G', 1, 1],
        ],
        'NHL_SHOWDOWN': [
            ['FLEX', 6, 6],
        ],
        'MLB_SHOWDOWN': [
            ['CPT', 1, 1],
            ['FLEX', 5, 5],
        ],
        'XFL': [
            ['QB', 1, 1],
            ['RB', 1, 3],
            ['WR', 2, 4],
            ['DST', 1, 1]
        ],
        'TEN': [
            ['P', 6, 6],
        ],
        'PGA': [
            ['G', 6, 6],
        ],
        'CSGO_SHOWDOWN': [
            ['CPT', 1, 1],
            ['FLEX', 5, 5],
        ],
    },
    FAN_DUEL: {
        'NBA': [
            ['PG', 2, 2],
            ['SG', 2, 2],
            ['SF', 2, 2],
            ['PF', 2, 2],
            ['C', 1, 1],
        ],
        'MLB': [
            ['P', 1, 1],
            ['1B', 1, 2],  # TODO - allow C or 1B
            ['2B', 1, 2],
            ['3B', 1, 2],
            ['SS', 1, 2],
            ['OF', 3, 4],
        ],
        'WNBA': [
            ['G', 3, 3],
            ['F', 4, 4],
        ],
        'NFL': get_nfl_positions(d_abbrev='D'),
        'NFL_MVP': get_nfl_showdown_positions(fd=True),
        'NASCAR': [
            ['D', 5, 5],
        ],
        'PGA': [
            ['G', 6, 6],
        ],
    }
}

NBA_GENERAL_POSITIONS = [
    ['G', 3, 4],
    ['F', 3, 4],
    ['C', 1, 2],
]
MLB_GENERAL_POSITIONS = [
    ['P', 2, 2],
]
WNBA_GENERAL_POSITIONS = [
    ['G', 2, 3],
    ['F', 3, 4],
]


class RuleSet(object):
    def __init__(self, site, league,
                 roster_size, position_limits,
                 salary_max, salary_min=0,
                 general_position_limits=None,
                 offensive_positions=None, defensive_positions=None,
                 max_players_per_team=None,
                 min_teams=None,
                 position_per_team_rules=None,
                 game_type='classic'):
        self.site = site
        self.league = league
        self.roster_size = roster_size
        self.position_limits = position_limits
        self.general_position_limits = general_position_limits
        self.salary_min = salary_min
        self.salary_max = salary_max
        self.offensive_positions = offensive_positions
        self.defensive_positions = defensive_positions
        self.game_type = game_type
        self.max_players_per_team = max_players_per_team or (roster_size - 1)
        self.position_per_team_rules = position_per_team_rules
        self.min_teams = min_teams


DK_NBA_RULE_SET = RuleSet(
    site=DRAFT_KINGS,
    league='NBA',
    roster_size=ROSTER_SIZE[DRAFT_KINGS]['NBA'],
    salary_max=SALARY_CAP[DRAFT_KINGS]['NBA'],
    position_limits=POSITIONS[DRAFT_KINGS]['NBA'],
    general_position_limits=NBA_GENERAL_POSITIONS,
)

DK_NBA_SHOWDOWN_RULE_SET = RuleSet(
    site=DRAFT_KINGS,
    league='NBA_SHOWDOWN',
    roster_size=ROSTER_SIZE[DRAFT_KINGS]['NBA_SHOWDOWN'],
    salary_max=SALARY_CAP[DRAFT_KINGS]['NBA_SHOWDOWN'],
    position_limits=POSITIONS[DRAFT_KINGS]['NBA_SHOWDOWN'],
    game_type='showdown',
    general_position_limits=[],
)

FD_NBA_RULE_SET = RuleSet(
    site=FAN_DUEL,
    league='NBA',
    roster_size=ROSTER_SIZE[FAN_DUEL]['NBA'],
    salary_max=SALARY_CAP[FAN_DUEL]['NBA'],
    position_limits=POSITIONS[FAN_DUEL]['NBA'],
    general_position_limits=NBA_GENERAL_POSITIONS,
)

DK_WNBA_RULE_SET = RuleSet(
    site=DRAFT_KINGS,
    league='WNBA',
    roster_size=ROSTER_SIZE[DRAFT_KINGS]['WNBA'],
    salary_max=SALARY_CAP[DRAFT_KINGS]['WNBA'],
    position_limits=POSITIONS[DRAFT_KINGS]['WNBA'],
    general_position_limits=NBA_GENERAL_POSITIONS,
)

FD_WNBA_RULE_SET = RuleSet(
    site=FAN_DUEL,
    league='WNBA',
    roster_size=ROSTER_SIZE[FAN_DUEL]['WNBA'],
    salary_max=SALARY_CAP[FAN_DUEL]['WNBA'],
    position_limits=POSITIONS[FAN_DUEL]['WNBA'],
    general_position_limits=NBA_GENERAL_POSITIONS,
)

DK_NFL_RULE_SET = RuleSet(
    site=DRAFT_KINGS,
    league='NFL',
    roster_size=ROSTER_SIZE[DRAFT_KINGS]['NFL'],
    salary_max=SALARY_CAP[DRAFT_KINGS]['NFL'],
    position_limits=POSITIONS[DRAFT_KINGS]['NFL'],
    offensive_positions=['QB', 'RB', 'WR', 'TE'],
    defensive_positions=['DST'],
    general_position_limits=[],
)

FD_NFL_RULE_SET = RuleSet(
    site=FAN_DUEL,
    league='NFL',
    roster_size=ROSTER_SIZE[FAN_DUEL]['NFL'],
    salary_max=SALARY_CAP[FAN_DUEL]['NFL'],
    position_limits=POSITIONS[FAN_DUEL]['NFL'],
    offensive_positions=['QB', 'RB', 'WR', 'TE'],
    defensive_positions=['D'],
    general_position_limits=[],
)

DK_NFL_SHOWDOWN_RULE_SET = RuleSet(
    site=DRAFT_KINGS,
    league='NFL_SHOWDOWN',
    roster_size=ROSTER_SIZE[DRAFT_KINGS]['NFL_SHOWDOWN'],
    salary_max=SALARY_CAP[DRAFT_KINGS]['NFL_SHOWDOWN'],
    position_limits=POSITIONS[DRAFT_KINGS]['NFL_SHOWDOWN'],
    offensive_positions=['CPT'],
    defensive_positions=['DST'],
    general_position_limits=[],
    game_type='showdown',
)

FD_NFL_MVP_RULE_SET = RuleSet(
    site=FAN_DUEL,
    league='NFL_SHOWDOWN',
    roster_size=ROSTER_SIZE[FAN_DUEL]['NFL_MVP'],
    salary_max=SALARY_CAP[FAN_DUEL]['NFL_MVP'],
    position_limits=POSITIONS[FAN_DUEL]['NFL_MVP'],
    offensive_positions=['CAPT'],
    defensive_positions=['D'],
    general_position_limits=[],
    game_type='showdown',
)

FD_PGA_RULE_SET = RuleSet(
    site=FAN_DUEL,
    league='PGA',
    roster_size=ROSTER_SIZE[FAN_DUEL]['PGA'],
    salary_max=SALARY_CAP[FAN_DUEL]['PGA'],
    position_limits=POSITIONS[FAN_DUEL]['PGA'],
    general_position_limits=[],
)

DK_PGA_RULE_SET = RuleSet(
    site=DRAFT_KINGS,
    league='PGA',
    roster_size=ROSTER_SIZE[DRAFT_KINGS]['PGA'],
    salary_max=SALARY_CAP[DRAFT_KINGS]['PGA'],
    position_limits=POSITIONS[DRAFT_KINGS]['PGA'],
    general_position_limits=[],
)

FD_NASCAR_RULE_SET = RuleSet(
    site=FAN_DUEL,
    league='NASCAR',
    roster_size=ROSTER_SIZE[FAN_DUEL]['NASCAR'],
    salary_max=SALARY_CAP[FAN_DUEL]['NASCAR'],
    position_limits=POSITIONS[FAN_DUEL]['NASCAR'],
    general_position_limits=[],
)

DK_MLB_RULE_SET = RuleSet(
    site=DRAFT_KINGS,
    league='MLB',
    # Can start two pitchers and 5 hitters
    max_players_per_team=7,
    roster_size=ROSTER_SIZE[DRAFT_KINGS]['MLB'],
    salary_max=SALARY_CAP[DRAFT_KINGS]['MLB'],
    position_limits=POSITIONS[DRAFT_KINGS]['MLB'],
    general_position_limits=MLB_GENERAL_POSITIONS,
    min_teams=3,
    position_per_team_rules=[[lambda pos: 'P' not in pos, 5]],
)

FD_MLB_RULE_SET = RuleSet(
    site=FAN_DUEL,
    league='MLB',
    roster_size=ROSTER_SIZE[FAN_DUEL]['MLB'],
    salary_max=SALARY_CAP[FAN_DUEL]['MLB'],
    position_limits=POSITIONS[FAN_DUEL]['MLB'],
    general_position_limits=[],
)

DK_SOCCER_RULE_SET = RuleSet(
    site=DRAFT_KINGS,
    league='SOCCER',
    roster_size=ROSTER_SIZE[DRAFT_KINGS]['SOCCER'],
    salary_max=SALARY_CAP[DRAFT_KINGS]['SOCCER'],
    position_limits=POSITIONS[DRAFT_KINGS]['SOCCER'],
    offensive_positions=['M', 'F'],
    defensive_positions=['GK', 'D'],
    general_position_limits=[],
)

DK_EURO_LEAGUE_RULE_SET = RuleSet(
    site=DRAFT_KINGS,
    league='EL',
    roster_size=ROSTER_SIZE[DRAFT_KINGS]['EL'],
    salary_max=SALARY_CAP[DRAFT_KINGS]['EL'],
    position_limits=POSITIONS[DRAFT_KINGS]['EL'],
    general_position_limits=[],
)

DK_NHL_RULE_SET = RuleSet(
    site=DRAFT_KINGS,
    league='NHL',
    roster_size=ROSTER_SIZE[DRAFT_KINGS]['NHL'],
    salary_max=SALARY_CAP[DRAFT_KINGS]['NHL'],
    position_limits=POSITIONS[DRAFT_KINGS]['NHL'],
    offensive_positions=['C', 'W'],
    defensive_positions=['G', 'D'],
    general_position_limits=[],
)

DK_NHL_SHOWDOWN_RULE_SET = RuleSet(
    site=DRAFT_KINGS,
    league='NHL_SHOWDOWN',
    roster_size=ROSTER_SIZE[DRAFT_KINGS]['NHL_SHOWDOWN'],
    salary_max=SALARY_CAP[DRAFT_KINGS]['NHL_SHOWDOWN'],
    position_limits=POSITIONS[DRAFT_KINGS]['NHL_SHOWDOWN'],
    offensive_positions=['C', 'W'],
    defensive_positions=['G', 'D'],
    general_position_limits=[],
    game_type='showdown',
)

DK_NBA_PICKEM_RULE_SET = RuleSet(
    site=DRAFT_KINGS,
    league='NBA',
    roster_size=6,
    salary_max=None,
    position_limits=None,
    game_type='pickem',
)

DK_MLB_SHOWDOWN_RULE_SET = RuleSet(
    site=DRAFT_KINGS,
    league='MLB_SHOWDOWN',
    roster_size=ROSTER_SIZE[DRAFT_KINGS]['MLB_SHOWDOWN'],
    salary_max=SALARY_CAP[DRAFT_KINGS]['MLB_SHOWDOWN'],
    position_limits=POSITIONS[DRAFT_KINGS]['MLB_SHOWDOWN'],
    general_position_limits=[],
    game_type='showdown'
)


DK_XFL_CLASSIC_RULE_SET = RuleSet(
    site=DRAFT_KINGS,
    league='XFL',
    roster_size=ROSTER_SIZE[DRAFT_KINGS]['XFL'],
    salary_max=SALARY_CAP[DRAFT_KINGS]['XFL'],
    position_limits=POSITIONS[DRAFT_KINGS]['XFL'],
    general_position_limits=[],
)


DK_TEN_CLASSIC_RULE_SET = RuleSet(
    site=DRAFT_KINGS,
    league='TEN',
    roster_size=ROSTER_SIZE[DRAFT_KINGS]['TEN'],
    salary_max=SALARY_CAP[DRAFT_KINGS]['TEN'],
    position_limits=POSITIONS[DRAFT_KINGS]['TEN'],
    general_position_limits=[],
)

DK_CSGO_SHOWDOWN = RuleSet(
    site=DRAFT_KINGS,
    league='CSGO_SHOWDOWN',
    roster_size=ROSTER_SIZE[DRAFT_KINGS]['CSGO_SHOWDOWN'],
    salary_max=SALARY_CAP[DRAFT_KINGS]['CSGO_SHOWDOWN'],
    position_limits=POSITIONS[DRAFT_KINGS]['CSGO_SHOWDOWN'],
    game_type='showdown',
    general_position_limits=[],
    max_players_per_team=3,
)



from draftfast.orm import NFLRoster, Player, ShowdownRoster
from nose import tools as ntool


def test_roster_equality():
    player_a = Player(pos='RB', name='A', cost=1, team='X')
    player_b = Player(pos='QB', name='B', cost=1, team='X')
    player_c = Player(pos='QB', name='C', cost=1, team='X')

    roster_a = NFLRoster()
    roster_a.add_player(player_a)
    roster_a.add_player(player_b)

    roster_b = NFLRoster()
    roster_b.add_player(player_a)
    roster_b.add_player(player_b)

    roster_c = NFLRoster()
    roster_c.add_player(player_a)
    roster_c.add_player(player_c)

    ntool.assert_false(roster_a == roster_c)
    ntool.assert_true(roster_a == roster_b)


def test_roster_equality_with_position_shuffle():
    player_a = Player(pos='RB', name='A', cost=1, team='X')
    player_a_new_pos = Player(pos='WR', name='A', cost=1, team='X')
    player_b = Player(pos='QB', name='B', cost=1, team='X')
    player_c = Player(pos='QB', name='C', cost=1, team='X')

    roster_a = NFLRoster()
    roster_a.add_player(player_a)
    roster_a.add_player(player_b)

    roster_b = NFLRoster()
    roster_b.add_player(player_a_new_pos)
    roster_b.add_player(player_b)

    roster_c = NFLRoster()
    roster_c.add_player(player_a)
    roster_c.add_player(player_c)

    ntool.assert_false(roster_a == roster_c)
    ntool.assert_true(roster_a == roster_b)


def test_showdown_roster_equality_and_position_shuffle():
    """
    In Showdown, point changes result from CPT assignment,
    so two lineups are not equal given the same player.
    """
    player_a = Player(pos='FLEX', name='A', cost=1, team='A')
    player_a_new_pos = Player(pos='CPT', name='A', cost=1, team='A')
    player_b = Player(pos='FLEX', name='B', cost=1, team='C')
    player_c = Player(pos='FLEX', name='C', cost=1, team='D')

    roster_a = ShowdownRoster()
    roster_a.add_player(player_a)
    roster_a.add_player(player_b)

    roster_b = ShowdownRoster()
    roster_b.add_player(player_a_new_pos)
    roster_b.add_player(player_b)

    roster_c = ShowdownRoster()
    roster_c.add_player(player_a)
    roster_c.add_player(player_c)

    ntool.assert_false(roster_a == roster_c)
    ntool.assert_false(roster_a == roster_b)


def test_roster_set():
    player_a = Player(pos='RB', name='A', cost=1, team='X')
    player_b = Player(pos='QB', name='B', cost=1, team='X')
    player_c = Player(pos='QB', name='C', cost=1, team='X')

    roster_a = NFLRoster()
    roster_a.add_player(player_a)
    roster_a.add_player(player_b)

    roster_b = NFLRoster()
    roster_b.add_player(player_a)
    roster_b.add_player(player_b)

    roster_c = NFLRoster()
    roster_c.add_player(player_a)
    roster_c.add_player(player_c)
    ntool.assert_true(len(set([roster_a, roster_b, roster_c])), 2)



from draftfast.orm import Player
from terminaltables import AsciiTable

T1 = 'T1'
T2 = 'T2'
T3 = 'T3'
T4 = 'T4'
T5 = 'T5'
T6 = 'T6'

TIERS = [
    T1,
    T2,
    T3,
    T4,
    T5,
    T6,
]


class TieredLineup(object):

    def __init__(self, players):
        assert len(players) == 6, 'Must have six players'
        for idx, p in enumerate(players):
            if p.tier != TIERS[idx]:
                raise Exception(
                    'Player {} is not in tier {} but in tier {}'.format(
                        p.name, TIERS[idx], p.tier
                    )
                )

        [
            self.T1,
            self.T2,
            self.T3,
            self.T4,
            self.T5,
            self.T6,
        ] = players

    def __repr__(self):
        table_data = [[
            'Name',
            'Tier',
            'Matchup',
            'Projected',
            'vs Avg',
            'Locked',
        ]]
        for p in self.players:
            table_data.append(p.to_table_row())
        return (
            AsciiTable(table_data).table +
            '\n' +
            'Total: {}'.format(self.total)
        )

    @property
    def players(self):
        return [
            self.T1,
            self.T2,
            self.T3,
            self.T4,
            self.T5,
            self.T6,
        ]

    @property
    def total(self):
        return sum([p.proj for p in self.players])


class TieredPlayer(Player):

    def __init__(self, tier, **kwargs):
        self.tier = tier
        super(TieredPlayer, self).__init__(
            **kwargs
        )

    def to_table_row(self):
        return [
            self.name,
            self.tier,
            self.matchup,
            self.proj,
            self.v_avg,
            'LOCK' if self.lock else '',
        ]



import os
import subprocess
import csv
from itertools import islice

from draftfast import dke_exceptions as dke

upload_file = '{}/data/current-upload.csv'.format(os.getcwd())


def create_upload_file():
    subprocess.call(['touch', upload_file])
    with open(upload_file, 'w') as f:
        writer = csv.writer(f)
        writer.writerow(
            [
                'P',
                'P',
                'C',
                '1B',
                '2B',
                '3B',
                'SS',
                'OF',
                'OF',
                'OF',
            ])


def map_pids(pid_file):
    player_map = {}
    with open(pid_file, 'r') as f:
        n = 0
        fields = None
        for line in f.readlines():
            n += 1
            if 'TeamAbbrev' in line:  # line with field names was found
                fields = line.split(',')
                break

        if not fields:
            raise dke.InvalidCSVUploadFileException(
                "Check that you're using the DK CSV upload template, " +
                "which can be found at " +
                "https://www.draftkings.com/lineup/upload.")

        f.close()
        f = islice(open(pid_file, 'r'), n, None)
        reader = csv.DictReader(f, fieldnames=fields)
        for line in reader:
            player_map[line['Name'] + ' ' + line['Position']] = line['ID']

    return player_map


def update_upload_csv(player_map, roster):
    sorted_players = roster.sorted_players()
    with open(upload_file, 'a') as f:
        writer = csv.writer(f)
        writer.writerow([
            p.get_player_id(player_map)
            for p in sorted_players
        ])
